/*
  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  |           =====>>>>> 2002 <<<<=====             |
  |                                                 |
  \-------------------------------------------------/
  
  ===================================================
  
  Name:       ccm.C
  
  @Copyright: Damian J. J. Farnell
  
  Date:       Wed Sep 25 15:32:06 BST 2002
  
  Purpose:    Determine and solve CCM equations.
  
  ===================================================
  
  ---------------------------------------------------

  MODULE NAME: CCM.C 

  PURPOSE:     MAIN ROUTINE WHICH CALLS EACH MODULE

  ---------------------------------------------------

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// #include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#if USE_SHM  /* use SHM to store matrix on clusters with slow or no disk */
#if defined(sparc)
#define _POSIX_C_SOURCE 3  /* needed for shm_open @SunOS.mman.h */
#endif
#include <unistd.h> /* close */
// #include <sys/file.h> /* O_RDWR */
#include <sys/fcntl.h> /* O_RD_WR */
#include <sys/mman.h>
#endif

#include "../include/nr.h"
#include "../include/nrutil.h"
#include "../include/Constants.h"
#include "../include/GlobalVariables.h"
#include "../include/FunctionPrototypes.h"

#if (PARALLEL_FLAG)
#include<mpi.h>
#endif

// Run in parallel with:

// mpirun -n 1 -p ./ccm.x 

main(int argc, char *argv[])
{
  
  int i;
  int run_option;
  time_t  t1, t2;   /* measure speed */

#if (PARALLEL_FLAG)

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD,&npe);
  MPI_Comm_rank(MPI_COMM_WORLD,&myid);

  if (npe > 16) {
    printf("Something wrong in ccm.C, npe > 16...\n"
           "Exiting!!\n\n");
    exit(1);
  }
  printf("NPE = %d MYID = %d\n", npe, myid);

#else

  npe=1;
  myid=0;

#endif

  /* ccm.C: 

     Purpose: To enumerate all CCM spin cluster configurations within 
     an LSUBm (or other) approach and then determine and solve the
     CCM ground ket- and bra-state equations.

     ====> The Code Is Split Into 5 Modules: <====

     1) Input.C

     ..(a) Read In Control Variables and Initialise Data.

     ..(b) Define the crystallographic lattice.

     2) Define.C

     ..(a) Enumerate the connected clusters.

     ..(b) Determine the unconnected clsuters.

     ..(c) Choose a subset of these clusters (e.g., those
     ......which conserve s_T^z).

     3) Obtain.C

     ..(a) Loop over total number of fundamental configs.
     
     ..(b) Determine the CCM equations for each `term.'
  
     ..(c) Save CCM equations to disk.

     4. Collect.C

     ..(a) Collects like terms together and thus reduces disk usage.

     5. Solve.C

     ..(a) Solve ket- and bra-state equations.
     ..(b) Find ground-state energy, sublattice magnetisation, etc.

  */

  time(&t1);        /* get unix time in seconds */

  terms_fp=NULL;

  if (argc == 3) {
    run_option=atoi(argv[2]);
    if (myid==0)
      printf("Run option is =%d\n",run_option);
  } else {
    printf("Input script name and run option...\nExiting!!\n\n");
    exit(1);
  }

  fflush(stdout);
  /* dont buffer output to get all outputs immediately */
  setvbuf(stdout,NULL,_IONBF,0);
  /* be sure all threads have runoption set */
#if (PARALLEL_FLAG) 
  MPI_Barrier(MPI_COMM_WORLD);
#endif

  input(argc,argv);

#if (PARALLEL_FLAG) 
  MPI_Barrier(MPI_COMM_WORLD);
#endif

  if (run_option == 1 || run_option == 4) 
    define();

#if (PARALLEL_FLAG) 
  MPI_Barrier(MPI_COMM_WORLD);
#endif
  time(&t2);        /* get unix time in seconds */
  printf("Time: define finished after %ds\n",t2-t1);

  /* store equations to file (slow disk?) or shared memory (SHM) */
#if USE_SHM  /* use SHM to store matrix on clusters with slow or no disk */
             /* its used as memory file system, nothing is shared */
  if (run_option > 1) {
    int oflag=0;
    if (run_option == 2 || run_option == 4) { oflag=O_RDWR|O_CREAT|O_TRUNC; }
                                       else { oflag=O_RDONLY; }
    /* typo in linux shm_open manpage: O_RDWR versus O_RWDR */
    /* terms_fname[80] terms_shm_fd=-1 terms_fp=NULL */
    /* Warning, other processes with same name? */
    sprintf(terms_fname,"/ccm_terms_%d",myid); /* slash + 14 chars recommended */
#ifdef O_LARGEFILE
    oflag|=O_LARGEFILE; /* not in SUNOS-5.8 man page of shm_open, try pfiles */
#else
    if (myid==0) printf("Warning: no O_LARGEFILE, filelimit is 2GB\n");
#endif
    terms_shm_fd = shm_open(terms_fname, oflag, (6*8+0)*8+0); /* perm=00600 */
    if (terms_shm_fd<0) {
      printf("%d Error: shm_open(\"%s\") failed\n",myid,terms_fname);
      exit(1);
    }
    if (run_option == 2 || run_option == 4) 
           terms_fp = fdopen( terms_shm_fd, "w+" );
    else   terms_fp = fdopen( terms_shm_fd, "r" );
    if (!terms_fp) {
      printf("%d: fdopen(%d=%s) failed",myid,terms_shm_fd,terms_fname);
      exit(1);
    }
    if(myid==0) printf("\nThe term file is opened in shared memory as %s.\n"
     "Please have in mind, that this may influence other processes.\n\n",
      terms_fname);
  }
#else /* use ordinary files, not recomended on SunOS or NFS-Clusters */
  if (run_option > 1) {
    terms_fp=NULL;
    sprintf(terms_fname,"%s.input",Control.name); /* ../Data/XXX.input */
    if (run_option == 2 || run_option == 4) 
           terms_fp = fopen( terms_fname, "w+" );
    else   terms_fp = fopen( terms_fname, "r" );
    if (!terms_fp) {
      printf("%d: fopen(\"%s\") failed\n",myid,terms_fname);
      exit(1);
    }
    fp_buf=(char *)malloc(1024*1024); /* 1MB IO-Buffer, minimize sys-time */
    if (fp_buf)
      setvbuf(terms_fp,fp_buf,_IOFBF,1024*1024);
    if(myid==0) printf("\nWARNING: The term file is opened on disk as %s.\n"
     "If there is enough local shared memory available, define USE_SHM\n"
     "to hold the ccm-equations in memory (often much faster).\n\n",
     terms_fname);
  }
#endif

  if (run_option == 2 || run_option == 4) {
    obtain();  /* get and store collected CCM terms to data.input */
    /* ToDo: obtain is badly balanced 
     *    we should start with slowest term and put the next easier term to
     *    the next free mpi-jobs until all jobs are done
     *    output 3 biggest and smalles number of terms per cluster? 
     */
  }
  /* "OBTAIN ... SUCCESSFULLY COMPLETED" outputted */
  time(&t2);        /* get unix time in seconds */
#if (PARALLEL_FLAG) 
  MPI_Barrier(MPI_COMM_WORLD);  /* 100% CPU, nice 19 */
#endif
  printf("Time myid= %d : obtain finished after %ds\n",myid,t2-t1);
#if (PARALLEL_FLAG) 
  MPI_Barrier(MPI_COMM_WORLD); /* wait for other outputs */
#endif
  

  if (run_option == 3 || run_option == 4) 
    solve();   /* read data.input (often) and output data.output */

  time(&t2);        /* get unix time in seconds */
#if (PARALLEL_FLAG)
  MPI_Finalize();
#endif
  printf("Time myid= %d : solve finished after %ds\n",myid,t2-t1);

  if (terms_fp) {
    if (fclose(terms_fp)) 
      printf("%d: Could not close term file\n",myid);
    terms_fp=0; 
  }
  if (fp_buf) free(fp_buf);
#if USE_SHM
  /* free shared memory */
  if (terms_shm_fd) {
    if (!close(terms_shm_fd)) 
      printf("%d: Could not close file descriptor of shm-object\n",myid);
    terms_shm_fd=0; 
  }
  if (run_option == 3 || run_option == 4) {
    if (shm_unlink(terms_fname))
      printf("%d: Warning shared memory (SHM) could not be released\n", myid);
  }
#endif

  // if (myid==0)
  //  printf("\nPress a key to close this window.\n\n");
  //  getchar();

}


